package Net::Bummskraut::Connection;
use strict;
use JSON::XS;
use IO::Socket::INET;
use Fcntl;
use POSIX;
use Net::Bummskraut::Event;
use Net::Bummskraut::Protocol;

our @ISA = qw/Net::Bummskraut::Event Net::Bummskraut::Protocol/;

sub new {
   my $this = shift;
   my $class = ref($this) || $this;
   my $self = { @_ };
   bless $self, $class;
   $self->init_protocol; # for Net::Bummskraut
   if (exists $self->{socket}) {
      binmode $self->{socket};
      _set_noblock ($self->{socket});
   }
   return $self;
}

sub _set_noblock {
   my ($s) = @_;
   my $flags = 0;
   fcntl($s, F_GETFL, $flags)
       or die "Couldn't get flags for HANDLE : $!\n";
   $flags |= O_NONBLOCK;
   fcntl($s, F_SETFL, $flags)
       or die "Couldn't set flags for HANDLE: $!\n";
}

sub connect {
   my ($self, $host, $port) = @_;

   $self->{socket}
      and return;

   my $sock = IO::Socket::INET->new (
      PeerAddr => $host,
      PeerPort => $port,
      Proto    => 'tcp',
      Blocking => 0,
   );

   $self->{socket} = $sock;
   $self->{host}   = $host;
   $self->{port}   = $port;

   binmode $sock;

   $self->{cw} =
      AnyEvent->io (poll => 'w', fh => $sock, cb => sub {
         if ($! = $sock->sockopt (SO_ERROR)) {
            $self->event ('connect_error' => $!);
            $self->disconnect ("Couldn't connect");
         } else {
            $self->start_reader;
            $self->start_writer;
            $self->event ('connect');
         }
         delete $self->{cw};
      });
}

sub disconnect {
   my ($self, $reason) = @_;
   delete $self->{cw};
   delete $self->{r};
   delete $self->{w};
   eval {
      $self->{socket}->close;
   };
   if (delete $self->{socket}) {
      $self->event (disconnect => $reason || "Disconnect without reason.");
   }
}

sub start_reader {
   my ($self) = @_;

   my ($host, $port, $sock) = ($self->{host}, $self->{port}, $self->{socket});

   $self->{r} =
      AnyEvent->io (poll => 'r', fh => $sock, cb => sub {
         my $l = sysread $sock, my $data, 1024;

         if (defined $l) {
            if ($l == 0) {
               $self->disconnect ("EOF from bummskraut_server '$host:$port'");
            } else {
               $self->{read_buffer} .= $data;
               $self->handle_data (\$self->{read_buffer});
            }

         } else {
            return if $! == EAGAIN();
            $self->disconnect (
               "Error while reading from bummskraut server '$host:$port': $!"
            );
         }
      });
}

sub start_writer {
   my ($self) = @_;
   return unless $self->{r};
   return unless length ($self->{write_buffer}) > 0;

   unless ($self->{w}) {
      $self->{w} =
         AnyEvent->io (poll => 'w', fh => $self->{socket}, cb => sub {
            my $data = $self->{write_buffer};

            if (defined ($data) && $data ne '') {
               my $len = syswrite $self->{socket}, $data;

               if (defined $len) {
                  if ($len == length $self->{write_buffer}) {
                     delete $self->{w};
                  }

                  $self->{write_buffer} = substr $self->{write_buffer}, $len;
               } else {
                  return if $! == EAGAIN();
                  $self->disconnect (
                     "Error when writing data on $self->{host}:$self->{port}: $!"
                  );
               }
            }
         });
   }
}

sub handle_data {
   my ($self, $buf) = @_;

   while ($$buf =~ m/^(\s*(\d+) )(.*)$/s) {
      my ($prefix, $len, $rembuf) = ($1, $2, $3);
      if ((length $rembuf) >= $len) {
         my $data = substr $rembuf, 0, $len;
         substr $$buf, 0, (length $prefix) + (length $data), '';
         $self->event (packet => from_json ($data));
      } else {
         return
      }
   }
}

sub send_data {
   my ($self, $data) = @_;
   $self->SUPER::send_data ($data);

   my $json = JSON::XS->new->utf8->allow_blessed (1);
   my $dump;
   $dump = $json->encode ($data);
   $self->write_data ((length $dump) . " " . $dump . "\015\012");
}

sub write_data {
   my ($self, $data) = @_;
   $self->{write_buffer} .= $data;
   $self->start_writer;
}

1;
