package Net::Bummskraut::Protocol;

use warnings;
use strict;

=head1 NAME

Net::Bummskraut::Protocol - A chat protocol based on JSON

=head1 VERSION

Version 0.01

=cut

our $VERSION = '0.01';

our %SCOPES;

=head1 SYNOPSIS

    package YourProtocolHandler;
    use Net::Bummskraut::Protocol;
    our @ISA = qw/Net::BUmmskraut::Protocol/;

=cut

sub init_protocol {
   my ($self) = @_;
   $self->reg_cb (packet => sub { $self->packet ($_[1]) });
}

=head1 DESCRIPTION

This modul is the documentation and implementation of the bummskraut
protocol. It uses JSON 

=head1 PROTOCOL

All messages from the client can have an _id field which
must be mirrored for a reply sent by the server.

=cut

sub pevent {
   my ($self, $ev, $packet, @arg) = @_;

   if (defined ($packet->{_id}) && $self->{cbs}->{$packet->{_id}}) {
      $self->{cbs}->{$packet->{_id}}->($self, $ev, $packet, @arg)
         and return;
   }

   $self->event ($ev, $packet, @arg);
}

sub packet {
   my ($self, $packet) = @_;

   if (my $sc = $SCOPES{$packet->{scope}}) {
      my $cmd = $sc->{$packet->{command}};
      $cmd->($self, $packet);
   }

   $self->event (debug_recv => $packet);
}

sub send_data {
   my ($self, $packet) = @_;
   $self->event (debug_send => $packet);
}

sub send_response {
   my ($self, $packet, $data) = @_;
   if (defined ($packet) && $packet->{_id}) {
      $data->{_id} = $packet->{_id};
   }
   $self->send_data ($data);
}

sub send_request {
   my ($self, $cb, $data) = @_;
   $data->{_id} = ++$self->{_id_cnt};
   $self->{cbs}->{$self->{_id_cnt}} = $cb
      if defined $cb;
   $self->send_data ($data);
}

=head2 Connection initiation

After connect the client sends:

   {
      "scope":"connection",
      "command":"hello",
      "auth":"<authmech>",
     ["response":"<authresponse>"]
   }

   Response: auth challenge
         or: hello

After that the server sends an auth-challenge.
If the authentication was successful the server also sends a hello.

After the server hello the server sends state changes
for the available ids.
If the <id> has any subids a subid add message is sent.

<authmech> := "custom-md5" (the only supported one currently)

=cut

sub srv_hello {
   my ($self, $packet) = @_;
   $self->send_response ($packet, { scope => "connection", command => "hello" });
}

sub cl_hello {
   my ($self, $cb) = @_;
   $self->send_request ($cb,
      { scope => "connection", command => "hello", "auth" => "custom-md5" }
   );
}

$SCOPES{connection}->{hello} = sub {
   my ($self, $packet) = @_;
   $self->pevent (hello => $packet);
};

=head2 Server auth challenge

   {
      "scope"     : "auth",
      "command"   : "challenge",
      "challenge" : "<authchallenge>"
   }

=cut

sub srv_auth_challenge {
   my ($self, $packet, $challenge) = @_;
   $self->send_response ($packet,
      { scope => "auth", command => "challenge", challenge => $challenge }
   );
}

$SCOPES{auth}->{challenge} = sub {
   my ($self, $packet) = @_;
   $self->pevent (auth_challenge => $packet, $packet->{challenge});
};

=head2 Client auth response

   {
      "scope"    : "auth",
      "command"  : "response",
      "response" : "<authresponse>"
   }

   Response: Auth challenge
         or: Hello

=cut

sub cl_auth_response {
   my ($self, $response, $cb) = @_;
   $self->send_request ($cb,
      { scope => "auth", command => "response", response => $response }
   );
}

$SCOPES{auth}->{response} = sub {
   my ($self, $packet) = @_;
   $self->pevent (auth_response => $packet, $packet->{response});
};

=head2 Configuration request to the server

   {
      "scope"     : "config",
      "configname": "<configname>",
      "command"   : "get",
   }

   Response: config state reply
         or: error

   {
      "scope"     : "config",
      "configname": "<configname>",
      "command"   : "set",
      "config"    : <json object>,
   }

   Response: info
         or: error

   <configname> := is a client unique key to store configuration data

=cut

sub cl_config_get {
   my ($self, $cfgname, $cb) = @_;

   $self->send_request ($cb,
      { scope => "config", configname => $cfgname, command => "get" }
   );
}

sub cl_config_set {
   my ($self, $cfgname, $cfg, $cb) = @_;

   $self->send_request ($cb,
      { scope => "config", configname => $cfgname, command => "set", config => $cfg }
   );
}

$SCOPES{config}->{get} = sub {
   my ($self, $packet) = @_;
   $self->pevent (config_get => $packet, $packet->{configname});
};

$SCOPES{config}->{set} = sub {
   my ($self, $packet) = @_;
   $self->pevent (config_set => $packet, $packet->{configname}, $packet->{config});
};

=head2 Configuration state from server

Used to send config state updates to clients.

   {
      "scope"     : "config",
      "configname": "<configname>",
      "command"   : "state",
      "config"    : <json object>,
   }

=cut

sub srv_config {
   my ($self, $packet, $cfgname, $cfg) = @_;

   $self->send_response ($packet,
      { scope => "config", configname => $cfgname, command => "state", config => $cfg }
   );
}

$SCOPES{config}->{state} = sub {
   my ($self, $packet) = @_;
   $self->pevent (config_state => $packet, $packet->{configname}, $packet->{config});
};

=head2 State changes from server and from client

If an ID's state changes following command is issued:

   {
      "scope"     : "id",
      "id"        : "<id>",
      "command"   : "state",
      "timestamp" : "<unixtimestamp>",
      "state"     : { "state":"<availstate>", "desc":"<desc>" },
      "src_alias" : <srcalias>
   }

   <id>         := an URI
   <availstate> := "present" | "chat" | "dnd" | "away" | "xa" | "n/a" | "dead"
   <desc>       := any string
   <srcalias>   := "<nickname>"  (sets the nickname)
                 | ""            (unsets the nickname)
                 | undef/null    (lets the nickname unchanged)

   On client request following responses must be sent: state change
   Or: error

=cut

sub srv_state_change {
   my ($self, $id, $ts, $availstate, $desc, $alias) = @_;

   $self->send_data ({
      scope => "id", command => "state", id => $id,
      timestamp => $ts,
      state => {
         state => "$availstate",
         desc  => "$desc",
      },
      alias => $alias
   });
}

$SCOPES{id}->{state} = sub {
   my ($self, $packet) = @_;

   $self->pevent (state_change => $packet,
      $packet->{id},
      $packet->{timestamp},
      $packet->{state}->{state},
      $packet->{state}->{desc},
      $packet->{alias}
   );
};

=head2 Sub id add from server

   {
      "scope"     : "subid",
      "id"        : "<id>",
      "command"   : "add",
      "timestamp" : "<unixtimestamp>",
      "subids"       : [ { "id":"<id>", "nick":"<nickname>" }, ... ]
   }

=cut

sub srv_add_subid {
   my ($self, $id, $timestamp, @subids) = @_;

   $self->send_data ({
      scope => 'subid', command => 'add', id => $id,
      timestamp => $timestamp,
      subids => [ map { { id => $_->[0], nick => $_->[1] } } @subids ],
   });
}

$SCOPES{subid}->{add} = sub {
   my ($self, $packet) = @_;

   $self->pevent (add_subid => $packet,
      $packet->{id}, $packet->{timestamp},
      [ map { [$_->{id}, $_->{nick}] } @{$packet->{subids}} ]
   );
};

=head2 Sub id remove from server

   {
      "scope"     : "subid",
      "id"        : "<id>",
      "command"   : "remove",
      "timestamp" : "<unixtimestamp>",
      "subids"    : [ { "id":"<id>", "nick":"<nickname>" }, ... ]
   }

=cut

sub srv_remove_subid {
   my ($self, $id, $timestamp, @subids) = @_;

   $self->send_data ({
      scope => 'subid', command => 'remove', id => $id,
      timestamp => $timestamp,
      subids => [ map { { id => $_->[0], nick => $_->[1] } } @subids ]
   });
}

$SCOPES{subid}->{remove} = sub {
   my ($self, $packet) = @_;

   $self->pevent (remove_subid => $packet,
      $packet->{id}, $packet->{timestamp},
      [ map { [$_->{id}, $_->{nick}] } @{$packet->{subids}} ]
   );
};

=head2 Subid change from server

   {
      "scope"     : "subid",
      "id"        : "<id>",
      "command"   : "change",
      "timestamp" : "<unixtimestamp>",
      "subid"     : { "id":"<id>", "nick":"<nickname>" },
      "new_subid" : { "id":"<id>", "nick":"<nickname>" },
   }

=cut

sub srv_change_subid {
   my ($self, $id, $timestamp, $subid, $newid, $oldnick, $newnick) = @_;

   $self->send_data ({
      scope => 'subid', command => 'change', id => $id,
      timestamp => $timestamp,
      subid     => { id => $subid, nick => $oldnick },
      new_subid => { id => $newid, nick => $newnick },
   });
}

$SCOPES{subid}->{change} = sub {
   my ($self, $packet) = @_;

   $self->pevent (change_subid => $packet,
      $packet->{id}, $packet->{timestamp},
      $packet->{subid}->{id},   $packet->{new_subid}->{id},
      $packet->{subid}->{nick}, $packet->{new_subid}->{nick}
   );
};

=head2 Messages from server

   {
      "scope"   : "id",
      "src"     : "<id>",
      "id"      : "<id>",
      "command" : "message",

      "message" : {
         "scope"     : "<msgscope>",
         "type"      : "<msgtype>",
         "is_echo"   : <true | false>,
         "highlight" : <true | false>,
         "timestamp" : "<unixtimestamp>",
         "content"   : "<msgcontent>"
      }
   }

   <msgscope>      := "private" | "public"
   <msgtype>       := "notice" | "normal" | "action"
   <unixtimestamp> := ordinary timestamp
   <msgcontent>    := any string

=head2 Message sent response from server

   {
      "scope"     : "id",
      "src"       : "<id>",
      "id"        : "<id>",
      "timestamp" : "<unixtimestamp>",
      "command"   : "message sent",
   }

=head2 Messages from client

   {
      "scope"   : "id",
      "id"      : "<id>",
      "command" : "message",

      "message" : {
         "scope"     : "<msgscope>",
         "type"      : "<msgtype>",
         "content"   : "<msgcontent>"
      }
   }

   Response: 'message sent' message.
         Or: Error from server
   Note: If the client sent the message an echo will be broadcasted to all clients.

=cut

sub srv_message_sent {
   my ($self, $packet, $src, $id, $ts) = @_;
   $self->send_response ($packet, {
      scope => 'id', command => 'message sent',
      src => $src, id => $id, timestamp => $ts,
   });
}

$SCOPES{id}->{'message sent'} = sub {
   my ($self, $packet) = @_;
   $self->pevent (message_sent => $packet, $packet->{src}, $packet->{id}, $packet->{timestamp});
};

sub srv_message {
   my ($self, $packet, $src, $id, $scope, $type, $is_echo,
       $do_highlight, $timestamp, $content) = @_;

   $self->send_data ({
      scope   => 'id', command => 'message',
      src     => $src,
      id      => $id,
      message => {
         scope     => $scope,
         type      => $type,
         is_echo   => $is_echo,
         highlight => $do_highlight,
         timestamp => $timestamp,
         content   => $content,
      }
   });
}

sub cl_message {
   my ($self, $id, $scope, $type, $content, $cb) = @_;

   $self->send_request ($cb, {
      scope => 'id', command => 'message', id => $id,
      message => { scope => $scope, type => $type, content => $content },
   });
}

$SCOPES{id}->{message} = sub {
   my ($self, $packet) = @_;

   $self->pevent (message => $packet,
      $packet->{src}, $packet->{id},
      $packet->{message}->{scope},
      $packet->{message}->{type},
      $packet->{message}->{is_echo},
      $packet->{message}->{highlight},
      $packet->{message}->{timestamp},
      $packet->{message}->{content},
   );
};

=head2 Meta messages from server

   {
      "scope"     : "id",
      "src"       : "<id>",
      "id"        : "<id>",
      "command"   : "meta message",
      "message"   : {
         "timestamp" : "<unixtimestamp>",
         "content"   : "<msgcontent>"
      }
   }

=cut

sub srv_meta_message {
   my ($self, $src, $id, $ts, $content) = @_;

   $self->send_data ({
      scope   => 'id',
      command => 'meta message',
      src     => $src,
      id      => $id,
      message => {
         timestamp => $ts,
         content   => $content,
      }
   });
}

$SCOPES{id}->{'meta message'} = sub {
   my ($self, $packet) = @_;

   $self->pevent (meta_message => $packet,
      $packet->{src}, $packet->{id},
      $packet->{message}->{timestamp},
      $packet->{message}->{content}
   );
};

=head2 Commands from client

=head3 Reload command

   {
      "scope"   : "connection",
      "command" : "reload",
   }

   Response: Info response
         Or: Error from server

=cut

sub cl_command_reload {
   my ($self, $cb) = @_;

   $self->send_request ($cb, { scope => 'connection', command => 'reload' });
}

$SCOPES{connection}->{reload} = sub {
   my ($self, $packet) = @_;
   $self->pevent (command_reload => $packet);
};

=head3 History command

   {
      "scope"   : "connection",
      "command" : "history",
      "id"      : "<id>",
      "lines"   : <number of log lines to send>
   }

   Response: Info response
         Or: Error from server

=cut

sub cl_command_history {
   my ($self, $cb, $id, $lines) = @_;

   $self->send_request ($cb, {
      scope => 'connection', command => 'history',
      id => $id,
      lines => $lines,
   });
}

$SCOPES{connection}->{history} = sub {
   my ($self, $packet) = @_;
   $self->pevent (command_history => $packet, $packet->{id}, $packet->{lines});
};

=head3 Subscribe client message

   {
      "scope"   : "id",
      "id"      : "<id>",
      "command" : "subscribe"
   }

   Response: Info response on success
             Error response on failure

=cut

sub cl_subscribe {
   my ($self, $id, $cb) = @_;

   $self->send_request ($cb, {
      scope => 'id', id => $id, command => 'subscribe'
   });
}

$SCOPES{id}->{subscribe} = sub {
   my ($self, $packet) = @_;

   $self->pevent (subscribe => $packet, $packet->{id});
};

=head3 Unsubscribe client message

   {
      "scope"   : "id",
      "id"      : "<id>",
      "command" : "unsubscribe"
   }

   Response: Info response on success
             Error response on failure

=cut

sub cl_unsubscribe {
   my ($self, $id, $cb) = @_;

   $self->send_request ($cb, {
      scope => 'id', id => $id, command => 'unsubscribe'
   });
}

$SCOPES{id}->{unsubscribe} = sub {
   my ($self, $packet) = @_;

   $self->pevent (unsubscribe => $packet, $packet->{id});
};

=head3 Set group client message

   {
      "scope"   : "id",
      "id"      : "<id>",
      "command" : "setgroup",
      "groups"  : [<groups>]
   }

   This can go to server to set the groups of an ID
   and alsofrom the server.

   reply from server: info that groups were set
                  or: error that groups weren't successfully set

=cut

sub cl_set_group {
   my ($self, $id, $groups, $cb) = @_;

   $self->send_request ($cb, {
      scope => 'id', id => $id, command => "setgroup",
      groups => [ @$groups ]
   });
}

sub srv_set_group {
   my ($self, $id, $groups) = @_;

   $self->send_data ({
      scope => 'id', id => $id, command => "setgroup",
      groups => [ @$groups ]
   });
}

$SCOPES{id}->{setgroup} = sub {
   my ($self, $packet) = @_;

   $self->pevent (set_group => $packet, $packet->{id}, $packet->{groups});
};

=head3 Set nick message from client to server

   {
      "scope" : "id",
      "id"    : "<id>",
      "command" : "setnick",
      "nick"  : "<nick>"
   }

   response: info or error

=cut

sub cl_set_nick {
   my ($self, $id, $nick, $cb) = @_;

   $self->send_request ($cb, {
      scope => 'id', id => $id, nick => $nick, command => 'setnick'
   });
}

$SCOPES{id}->{setnick} = sub {
   my ($self, $packet) = @_;

   $self->pevent (set_nick => $packet, $packet->{id}, $packet->{nick});
};

=head3 Set presence message from client to server

   {
      "scope" : "connection",
      "command" : "setpresence",
      "state" : { "state" : "<availstate>", "desc":"<desc>" }
   }

=cut

sub cl_set_presence {
   my ($self, $state, $desc, $cb) = @_;

   $self->send_request ($cb, {
      scope => 'connection', command => 'setpresence',
      state => { state => $state, desc => $desc }
   });
}


$SCOPES{connection}->{setpresence} = sub {
   my ($self, $packet) = @_;

   $self->pevent (set_presence => $packet, $packet->{state}->{state}, $packet->{state}->{desc});
};


=head3 Meta command

   {
      "scope"   : "id",
      "id"      : "<id>",
      "command" : "meta",
      "message" : "<data>",
   }

   Response: Info response
         Or: Error from server

This is a meta command for the underlying protocol.

=cut

sub cl_meta_message {
   my ($self, $id, $data, $cb) = @_;

   $self->send_request ($cb, {
      scope => 'id', command => 'meta',
      id => $id, message => $data,
   });
}

$SCOPES{id}->{meta} = sub {
   my ($self, $packet) = @_;

   $self->pevent (meta_command => $packet,
      $packet->{id}, $packet->{message}
   );
};

=head3 Raw protocol message

   {
      "scope"   : "id",
      "id"      : "<id>",
      "command" : "raw",
      "message" : "<data>"
   }

   Response: Info response
         Or: Error from server

=cut

sub cl_raw_message {
   my ($self, $id, $data, $cb) = @_;

   $self->send_request ($cb, {
      scope => 'id', command => 'raw',
      id => $id, message => $data,
   });
}

$SCOPES{id}->{raw} = sub {
   my ($self, $packet) = @_;

   $self->pevent (raw_command => $packet,
      $packet->{id}, $packet->{message}
   );
};

=head2 Info responses/messages from the server

   {
      "scope"   : "connection",
    [ "id"      : "<id>", ]
      "type"    : "<infotype>",
      "command" : "info",
      "message" : {
         "timestamp" : "<unixtimestamp>",
         "content"   : "<msgcontent>",
      }
   }

   <infotype> := "connected"  (connected successfully)
               | "subscribe"  (subscription was successful)

=cut

sub srv_info {
   my ($self, $packet, $id, $type, $timestamp, $content) = @_;

   my @id = $id ? (id => $id) : ();

   $self->send_response ($packet, {
      scope => 'connection', command => 'info', @id, type  => $type,
      message => { timestamp => $timestamp, content => $content },
   });
}

$SCOPES{connection}->{info} = sub {
   my ($self, $packet) = @_;

   $self->pevent (info => $packet,
      $packet->{id}, $packet->{type},
      $packet->{message}->{timestamp}, $packet->{message}->{content}
   );
};

=head2 Errors from the server

   {
      "scope"   : "id",
    [ "id"     : "<id>", ]
      "type"    : "<errortype>",
      "command" : "error",
      "message" : {
         "timestamp" : "<unixtimestamp>",
         "content"   : "<msgcontent>",
      }
   }

   <errortype> := "no_such_id"     (when the destination id couldn't be found)
                | "bad_scheme"     (when an id has an unkown scheme)
                | "subscribe"      (subscription failure)

=cut

sub srv_error {
   my ($self, $packet, $id, $type, $timestamp, $content) = @_;

   my @id = $id ? (id => $id) : ();

   $self->send_response ($packet, {
      scope => 'connection', command => 'error', @id, type  => $type,
      message => { timestamp => $timestamp, content => $content },
   });
}

$SCOPES{connection}->{error} = sub {
   my ($self, $packet) = @_;

   $self->pevent (error => $packet,
      $packet->{id}, $packet->{type},
      $packet->{message}->{timestamp}, $packet->{message}->{content}
   );
};

=head2 Client to Client Broadcasts

   {
      "scope"    : "broadcast",
      "command"  : "relay",
      "broadcast": {
         "timestamp"  : "<unixtimestamp>",
         "command"    : "<broadcast command>",
         "client_tag" : "<client tag>",
         "data"       : <json data>
      }
   }

   <broadcast command> := client specific broadcast command
   <client tag>        := client specific tag to give a clue
                          for which clients this broadcast has
                          relevance.
   <json data>         := any json data for this broadcast command

   Response: the sent broadcast will be
             relayed to the sender too (with the _id field
             set to the sent broadcast command)

=cut

sub cl_broadcast_relay {
   my ($self, $command, $tag, $data, $ts, $cb) = @_;

   $self->send_request ($cb, {
      scope   => 'broadcast',
      command => 'relay',
      broadcast => {
         command    => $command,
         client_tag => $tag,
         data       => $data,
         timestamp  => $ts,
      }
   });
}

sub srv_broadcast_relay {
   my ($self, $packet, $command, $tag, $data, $ts) = @_;

   $self->send_response ($packet, {
      scope   => 'broadcast',
      command => 'relay',
      broadcast => {
         command    => $command,
         client_tag => $tag,
         data       => $data,
         timestamp  => $ts,
      }
   });
}

$SCOPES{broadcast}->{relay} = sub {
   my ($self, $packet) = @_;

   $self->pevent (broadcast_relay => $packet,
      $packet->{broadcast}->{command},
      $packet->{broadcast}->{client_tag},
      $packet->{broadcast}->{data},
      $packet->{broadcast}->{timestamp}
   );
};

#=head2 Client to Server GUI message
#
#    {
#       "scope"    : "gui",
#       "command"  : "start",
#       "tag"      : "<guitag>",
#    }
#
#    {
#       "scope"    : "gui",
#       "command"  : "data"
#       "data"     : <guidata>
#    }
#
#=cut
#
#sub cl_gui_start {
#   my ($self, $tag) = @_;
#   $self->send_data ({ scope => 'gui', command => 'start', tag => $tag });
#}
#
#$SCOPES{gui}->{start} = sub {
#   my ($self, $packet) = @_;
#   $self->pevent (gui_start => $packet, $packet->{tag});
#};
#
#sub srv_gui_data {
#   my ($self, $data) = @_;
#   $self->send_data ({ scope => 'gui', command => 'data', data => $data });
#}
#
#sub cl_gui_data {
#   my ($self, $data) = @_;
#   $self->send_data ({ scope => 'gui', command => 'start', tag => $tag });
#}
#
#$SCOPES{gui}->{data} = sub {
#   my ($self, $packet) = @_;
#   $self->pevent (gui_data => $packet, $packet->{data});
#};


=head1 AUTHOR

Robin Redeker, C<< <elmex at ta-sa.org> >>

=head1 BUGS

Please report any bugs or feature requests to
C<bug-net-chat-json at rt.cpan.org>, or through the web interface at
L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=Net-Bummskraut>.
I will be notified, and then you'll automatically be notified of progress on
your bug as I make changes.

=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc Net::Bummskraut

You can also look for information at:

=over 4

=item * AnnoCPAN: Annotated CPAN documentation

L<http://annocpan.org/dist/Net-Bummskraut>

=item * CPAN Ratings

L<http://cpanratings.perl.org/d/Net-Bummskraut>

=item * RT: CPAN's request tracker

L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=Net-Bummskraut>

=item * Search CPAN

L<http://search.cpan.org/dist/Net-Bummskraut>

=back

=head1 ACKNOWLEDGEMENTS

=head1 COPYRIGHT & LICENSE

Copyright 2007 Robin Redeker, all rights reserved.

This program is free software; you can redistribute it and/or modify it
under the same terms as Perl itself.

=cut

1; # End of Net::Bummskraut
