package Net::Bummskraut::Scheme::IRC;
use strict;
use AnyEvent;
use Net::Bummskraut;
use Net::Bummskraut::Util;
use Net::IRC3::Client::Connection;
use Net::IRC3::Util qw/prefix_nick encode_ctcp/;
use Net::Bummskraut::Scheme;
use Encode;

our @ISA = qw/Net::Bummskraut::Scheme/;

sub new {
   my $this = shift;
   my $class = ref($this) || $this;
   my $self = {@_};
   bless $self, $class;
   $self->{server}->reg_scheme (irc => $self);
   $self->{cfg} = $self->{server}->get_my_cfg ('irc');
   $self->update_connections;

   $self->{cfg}->{nick} ||= '';
   $self->{cfg}->{user} ||= '';
   $self->{cfg}->{real} ||= '';
   $self->{cfg}->{highlight} ||= [];
   $self->{cfg}->{auto_greet} ||= [];
   $self->{cfg}->{away_parts} ||= [];
   $self->{cfg}->{server_passes} ||= {};
   $self->{cfg}->{update_interval} ||= 120;

   $self->start_timer;

   return $self;
}

sub start_timer {
   my ($self) = @_;

   $self->{update_timer} =
      AnyEvent->timer (
         after => $self->{cfg}->{update_interval} || 120,
         cb    => sub {
            $self->update_connections;
            $self->start_timer;
         }
      );
}

sub cfg_keys {
   my ($self) = @_;

   (
      [nick      => "The nickname to use on IRC connections"],
      [user      => "The username to use on IRC connections"],
      [real      => "The realname to use on IRC connections"],
      [highlight => "A list of highlighted words"],
      [server_passes   => "Server passwords"],
      [update_interval => "The interval in seconds to update the connection states and retry connects"],
      [auto_greet => "A list of IDs where automatic greetings should be done."],
      [away_parts => "A list of IDs which are being parted when you go away."]
   )
}

sub subscribe {
   my ($self, $id, $cb) = @_;

   unless (grep { $self->cmp_id ($_, $id) } @{$self->{cfg}->{subscriptions}}) {
      push @{$self->{cfg}->{subscriptions}}, $id;
   }

   $self->{server}->commit_cfg;

   $self->update_connections;
   $self->join_channels;
   $self->update_nick_states;

   $cb->(1, "subscribed '$id'");
}

sub unsubscribe {
   my ($self, $id, $cb) = @_;

   @{$self->{cfg}->{subscriptions}} =
      grep { not $self->cmp_id ($_, $id) } @{$self->{cfg}->{subscriptions}};

   $self->{server}->commit_cfg;

   $self->update_connections;
   $self->join_channels;
   $self->update_nick_states;

   $cb->(1, "unsubscribed '$id'");
}

sub subscriptions {
   my ($self) = @_;
   @{$self->{cfg}->{subscriptions}}
}

sub set_presence {
   my ($self, $state, $desc) = @_;

   if ($state eq 'present') {
      for (values %{$self->{conns}}) {
         $_->send_srv ('AWAY');
      }
   } else {
      for (values %{$self->{conns}}) {
         $_->send_srv (AWAY => $desc ne '' ? $desc : $state);
      }
   }
}

sub update_connections {
   my ($self) = @_;

   my %conn;
   for my $id (@{$self->{cfg}->{subscriptions}}) {
      my ($h, $p) = $self->get_id_host_port ($id);
      $conn{"$h:$p"} = [$h, $p, $id];
   }

   for my $hp (values %conn) {
      unless ($self->get_connection ($hp->[2])) {
         if (my $con = $self->get_pending_connection ($hp->[2])) {
            $con->disconnect;
            delete $self->{conns_pending}->{"$hp->[0]:$hp->[1]"};
         }

         $self->connect ($hp->[0], $hp->[1]);
      }
   }
}

sub refresh {
   my ($self) = @_;
   $self->update_connections;
   $self->{server}->distribute_info (undef, irc_refresh => "Refreshed IRC connections");
}

sub reload {
   my ($self) = @_;
   for ((values %{$self->{conns_pending}}), (values %{$self->{conns}})) {
      $_->disconnect ("user reloaded");
   }

   $self->{reload_timer} =
      AnyEvent->timer (after => 10, cb => sub {
         $self->update_connections;
         delete $self->{reload_timer}
      });
}

sub to_irc_url {
   my ($self, $host, $p, $targ) = @_;
   $p ||= 6667;
   join_irc_uri ($targ, $host . ($p != 6667 ? ":$p" : ""))
}

sub cmp_id {
   my ($self, $id1, $id2) = @_;
   $id1 = $self->normalize_id ($id1);
   $id2 = $self->normalize_id ($id2);
   $id1 eq $id2
}

sub normalize_id {
   my ($self, $id) = @_;
   my ($tar, $auth) = split_irc_uri ($id);
   my ($h, $p) = split /:/, $auth, 2;
   $p ||= 6667;
   $auth = ($h.($p != 6667 ? ":$p" : ""));
   join_irc_uri ($tar, $auth)
}

sub get_id_host_port {
   my ($self, $dest) = @_;
   my ($tar, $auth) = split_irc_uri ($dest);
   my ($h, $p) = split /:/, $auth, 2;
   $p ||= 6667;
   ($h, $p)
}

sub get_connection_or_error {
   my ($self, $dest) = @_;
   my $con = $self->get_connection ($dest);
   unless ($con) {
      $self->{server}->distribute_error (undef, no_such_id => "No connection to '$dest'");
   }
   $con
}

sub get_connection {
   my ($self, $dest) = @_;
   $self->{conns}->{(join ":", $self->get_id_host_port ($dest))};
}

sub get_pending_connection {
   my ($self, $dest) = @_;
   $self->{conns_pending}->{(join ":", $self->get_id_host_port ($dest))};
}

sub get_src {
   my ($self, $dest) = @_;
   my $con = $self->get_connection_or_error ($dest)
      or return;

   my ($tar, $auth) = split_irc_uri ($dest);
   join_irc_uri ($con->nick, $auth)
}

sub get_dest_target {
   my ($self, $dest) = @_;
   my ($target, $auth) = split_irc_uri ($dest);
   $target
}

sub filter_highlight {
   my ($self, $pc, $content) = @_;

   my $localnick = $pc->nick;
   my $hl = $content =~ /\Q$localnick\E/i;

   for (@{$self->{cfg}->{highlight} || []}) {
      $hl ||= $content =~ /\Q$_\E/i;
      $hl and last;
   }

   $hl
}

sub connect {
   my ($self, $host, $port) = @_;
   $port ||= 6667;

   my $pc = Net::IRC3::Client::Connection->new;
   $pc->reg_cb (
      debug_recv => sub {
         my ($pc, $msg) = @_;

         my $log =
            sprintf "IRC RECV{cmd: %s, params: %s, prefix: %s}",
            $msg->{command},
            (join ", ", @{$msg->{params} || []}),
            $msg->{prefix};

         $self->{server}->debug_log ($log);

         1
      },
      debug_send => sub {
         my ($pc, $prefix, $command, $trailing, @params) = @_;

         my $log =
            sprintf "IRC SEND{cmd: %s, params: %s, prefix: %s}",
            $command,
            (join ", ", @params, $trailing),
            $prefix;

         $self->{server}->debug_log ($log);

         1
      },
      publicmsg => sub {
         my ($pc, $chan, $msg) = @_;
         my ($from_nick) = Net::IRC3::Util::split_prefix ($msg->{prefix});
         my $src         = $self->to_irc_url ($host, $port, $from_nick);
         my $dst         = $self->to_irc_url ($host, $port, $chan);

         my $trail = decode ('utf8', $msg->{trailing});
         my $hl = $self->filter_highlight ($pc, $trail);

         $self->{server}->distribute_message (
            $src, $dst, 'public',
            uc ($msg->{command}) eq 'PRIVMSG' ? 'normal' : 'notice',
            0, $hl, time, $trail
         );
         1
      },
      privatemsg => sub {
         my ($pc, $chan, $msg) = @_;
         my $src = $self->to_irc_url ($host, $port, prefix_nick ($msg));

         # the multiple calls to this are not optimal, but they are
         # recognized by Server.pm and not sent to the client
         $self->{server}->change_state ($src, 'present', 'nick available', prefix_nick ($msg));

         $self->{server}->distribute_message (
            $src, $src, 'private',
            uc ($msg->{command}) eq 'PRIVMSG' ? 'normal' : 'notice',
            0, 0, time, decode ('utf8', $msg->{trailing})
         );
         1
      },
      irc_mode => sub {
         my ($pc, $ircmsg) = @_;
         my @p    = @{$ircmsg->{params}};
         my $who  = shift @p;
         my $src  = $self->to_irc_url ($host, $port, prefix_nick ($ircmsg));
         my $dst  = $self->to_irc_url ($host, $port, $who);
         my $mode = join " ", @p;
         $self->{server}->distribute_meta ($src, $dst, "mode change: $mode");

         1;
      },
      channel_add => sub {
         my ($pc, $chan, @nicks) = @_;
         my $dest = $self->to_irc_url ($host, $port, $chan);

         unless ($self->{channels}->{$dest}) {
            $self->{channels}->{$dest} = 1;
            $self->{server}->change_state ($dest, 'present', 'channel joined');

            if (grep { $self->cmp_id ($_, $dest) } @{$self->{cfg}->{auto_greet}}) {
               $self->{auto_greet_timer}->{$dest} =
                  AnyEvent->timer (after => rand (20), cb => sub {
                     my @msgs = qw/Hi hi hiho Yo!/;
                     $self->message (
                        $dest, public => 'normal', $msgs[int (rand (scalar @msgs))]
                     );
                     delete $self->{auto_greet_timer}->{$dest};
                  });
            }
         }

         $self->{server}->add_subids ($dest,
            map { [$self->to_irc_url ($host, $port, $_), $_] } @nicks
         );

         1
      },
      channel_remove => sub {
         my ($pc, $chan, @nicks) = @_;
         my $dest = $self->to_irc_url ($host, $port, $chan);

         if (grep { lc ($_) eq lc ($pc->nick) } @nicks) {
            delete $self->{channels}->{$dest};
            $self->{server}->change_state ($dest, 'n/a', 'channel left');

         } else {
            $self->{server}->remove_subids ($dest,
               map { [$self->to_irc_url ($host, $port, $_), $_] } @nicks
            );
         }

         1
      },
      channel_topic => sub {
         my ($pc, $chan, $topic, $who) = @_;
         $topic = decode ('utf8', $topic);
         my $dest = $self->to_irc_url ($host, $port, $chan);
         $who = $self->to_irc_url ($host, $port, $who) if defined $who;
         $self->{server}->distribute_meta ($who || $dest, $dest, "topic: $topic");

         1
      },
      channel_change => sub {
         my ($pc, $chan, $old_nick, $new_nick, $is_myself) = @_;

         my $src = $self->to_irc_url ($host, $port, $old_nick);
         my $dest = $self->to_irc_url ($host, $port, $chan);
         $self->{server}->change_subid (
            $dest,
            $src, $self->to_irc_url ($host, $port, $new_nick),
            $old_nick, $new_nick
         );

         $self->{server}->distribute_meta (
            $src, $dest, "nick change: $old_nick => $new_nick"
         );

         1
      },
      ctcp_VERSION => sub {
         my ($pc, $nick, $targ, $msg, $type) = @_;
         return 1 if $type ne 'PRIVMSG';
         $pc->send_srv (NOTICE => encode_ctcp (['VERSION', "Bummskraut:${Net::Bummskraut::VERSION}:Perl"]), $nick);
         1
   },
      ctcp_PING => sub {
         my ($pc, $nick, $targ, $msg, $type) = @_;
         return 1 if $type ne 'PRIVMSG';
         $pc->send_srv (NOTICE => encode_ctcp (['PING', $msg]), $nick);
         1
      },
      ctcp_ACTION => sub {
         my ($pc, $nick, $targ, $msg, $type) = @_;
         return 1 if $type ne 'PRIVMSG';
         my $trail = decode ('utf8', $msg);
         my $hl = $self->filter_highlight ($pc, $trail);

         my $src  = $self->to_irc_url ($host, $port, $nick);
         my $dest = $self->to_irc_url ($host, $port, $targ);

         if ($targ !~ m/^[!&+#]/) { # FIXME: broken IRC channel special case
            $dest = $src;
            # the multiple calls to this are not optimal, but they are
            # recognized by Server.pm and not sent to the client
            $self->{server}->change_state ($src, 'present', 'nick available', prefix_nick ($msg));
         }

         $self->{server}->distribute_message (
            $src, $dest, 'public', 'action', 0, $hl, time, $trail
         );

         1
      },
      ctcp => sub {
         my ($pc, $nick, $targ, $tag, $msg, $type) = @_;

         my $src  = $self->to_irc_url ($host, $port, $nick);
         my $dest = $self->to_irc_url ($host, $port, $targ);

         if ($targ !~ m/^[!&+#]/) { # FIXME: broken IRC channel special case
            $dest = $src;
         }

         if ($type eq 'PRIVMSG') {
            return 1 if $tag eq 'ACTION';
            $self->{server}->distribute_meta (
               $src, $dest, "CTCP[$tag]" . ($msg ne '' ? " $msg" : "")
            );
         } else {
            $self->{server}->distribute_meta (
               $src, $dest, "CTCP REPLY[$tag]" . ($msg ne '' ? " $msg" : "")
            );
         }

         1
      },
      registered => sub {
         my ($pc) = @_;

         $self->{conns}->{"$host:$port"} = $pc;
         delete $self->{conns_pending}->{"$host:$port"};
         $pc->enable_ping (90);
         $self->join_channels;
         $self->update_nick_states;

         my $dest = $self->to_irc_url ($host, $port, $pc->nick);
         $self->{server}->distribute_info (undef, irc_connect => "Connected to IRC Server $host:$port as '".$pc->nick."'");
         $self->{server}->change_state ($dest, 'present', '', $pc->nick);

         0
      },
      error => sub {
         my ($pc, $code, $msg, $ircmsg) = @_;
         my $n = Net::IRC3::Util::rfc_code_to_name ($code) || $code;
         $n =~ s/^ERR_//;

         my @p = @{$ircmsg->{params}}; pop @p; # pop message

         $self->{server}->distribute_error (
            undef, irc_error =>
                      "[$n]: $msg "
                      . (@p ? "(" . (join ", ", @p) . ")" : "")
         );
         1
      },
      disconnect => sub {
         $self->update_nick_states;
         delete $self->{conns}->{"$host:$port"};
         delete $self->{conns_pending}->{"$host:$port"};
         $self->{server}->distribute_info (undef, irc_disconnect => "Disconnected IRC Server $host:$port as '".$pc->nick."', reason: $_[1]");
         0
      },
   );

   eval {
      $pc->connect ($host, $port);
      $pc->register (
         $self->{cfg}->{nick} || "ncjson",
         $self->{cfg}->{user} || "ncjson",
         $self->{cfg}->{real} || "ncjson",
         $self->{cfg}->{server_passes}->{$host}
      );
      $self->{conns_pending}->{"$host:$port"} = $pc;
   };
   if ($@) {
      $self->{server}->distribute_error (undef, irc_connect_error => "Connection to '$host:$port' couldn't be established: $@");
   }
}

sub update_nick_states {
   my ($self) = @_;

   for my $id (@{$self->{cfg}->{subscriptions}}) {
      my $con = $self->get_connection ($id);
      if ($con) {
         my $d = $self->get_dest_target ($id);
         if ($d !~ m/^[!&+#]/) {
            $self->{server}->change_state ($id, 'present', 'nick available', $d);
         }

      } else {
         $self->{server}->change_state ($id, 'n/a', 'not connected');
      }
   }
}

sub join_channels {
   my ($self) = @_;

   # purge unsubscribed channels
   for my $con (values %{$self->{conns}}) {
      for my $c (keys %{$con->channel_list || {}}) {
         my ($h, $p) = ($con->{host}, $con->{port});
         my $cid     = $self->to_irc_url ($h, $p, $c);

         unless (grep { $self->cmp_id ($cid, $_) } @{$self->{cfg}->{subscriptions}}) {
            $con->send_srv (PART => "gone" => $c);
         }
      }
   }

   my %join;
   for my $id (@{$self->{cfg}->{subscriptions}}) {
      my $con = $self->get_connection ($id)
         or next;

      my ($target, $auth) = split_irc_uri ($id);
      $join{$auth}->[0] = $con;
      if ($target =~ m/^[&!+#]/ && not $con->channel_list ()->{lc $target}) {
         $join{$auth}->[1]->{$target} = 1;
      }
   }

   for (values %join) {
      my $con = $_->[0];
      my @chans = keys %{$_->[1]};
      $con->send_srv (JOIN => undef, join ",", @chans) if @chans;
   }
}

sub message {
   my ($self, $dest, $scope, $type, $content) = @_;
   my $src = $self->get_src ($dest)
      or return;
   my $con = $self->get_connection_or_error ($dest)
      or return;

   my $targ = $self->get_dest_target ($dest);

   $self->{server}->distribute_message (
      # FIXME: ^# is not sufficient to detect channel names!
      $src, $dest, ($targ =~ /^[&+#!]/ ? 'public' : 'private'),
      $type, 1, 0, my $ts = time, $content
   );

   $con->send_srv (($type eq 'normal' ? 'PRIVMSG' : 'NOTICE') => encode ('utf8', encode_ctcp ($content)) => $targ);
}

sub raw {
   my ($self, $dest, $msg) = @_;
   my $con;
   unless ($con = $self->get_connection ($dest)) {
      unless ($con = $self->get_pending_connection ($dest)) {
         $self->{server}->distribute_error (undef, no_such_id => "No connection to '$dest' in raw command.");
         return
      }
   }
   $con->send_raw ($msg);
}

sub meta {
   my ($self, $dest, $msg) = @_;
   my $src = $self->get_src ($dest)
      or return;
   my $con = $self->get_connection_or_error ($dest)
      or return;

   my $targ = $self->get_dest_target ($dest);

   if ($msg =~ /topic/) {
      $con->send_srv (TOPIC => undef, $targ);

   } elsif ($msg =~ /refresh/) {
      $self->update_connections;

   } elsif ($msg =~ /action\s*(.*?)\s*$/) {
      my $act = $1;
      $con->send_srv (PRIVMSG => encode ('utf8', encode_ctcp (['ACTION', $act])) => $targ);
      $self->{server}->distribute_message (
         # FIXME: ^# is not sufficient to detect channel names!
         $src, $dest, ($targ =~ /^[&+#!]/ ? 'public' : 'private'),
         'action', 1, 0, time, $act
      );

   } elsif ($msg =~ /version/) {
      $con->send_srv (PRIVMSG => encode ('utf8', encode_ctcp (['VERSION'])) => $targ);
   }
}

1;
