package Net::Bummskraut::Scheme::XMPP;
use strict;
use Net::Bummskraut::Util;
use Net::Bummskraut::Scheme;
use Net::XMPP2::Client;
use Net::XMPP2::Util qw/prep_bare_jid/;
use Net::Bummskraut::Scheme::XMPP::GUI;
use XML::Twig;

our @ISA = qw/Net::Bummskraut::Scheme/;

sub _dumpxml {
   my $data = shift;
   my $t = XML::Twig->new;
   if ($t->safe_parse ("<deb>$data</deb>")) {
      $t->set_pretty_print ('indented');
      return $t->sprint;
   } else {
      return "[$data]";
   }
}

sub new {
   my $this = shift;
   my $class = ref($this) || $this;
   my $self = {@_};
   bless $self, $class;

   $self->{server}->reg_scheme (xmpp => $self);
   $self->{cfg} = $self->{server}->get_my_cfg ('xmpp');
   $self->{cfg}->{jids} ||= {};
   $self->{cfg}->{via}  ||= {};
   $self->{cfg}->{nick} ||= "";
   $self->{cfg}->{main_jid} ||= "";
   $self->{cfg}->{update_interval} ||= 120;

   $self->{xmpp} = Net::XMPP2::Client->new;

   $self->init;

   $self->start_timer;

   $::GUI->{xmpp_gui} =
      Net::Bummskraut::Scheme::XMPP::GUI->new (scheme => $self);

   return $self;
}

sub start_timer {
   my ($self) = @_;

   $self->{update_timer} =
      AnyEvent->timer (
         after => $self->{cfg}->{update_interval} || 120,
         cb    => sub {
            $self->update_connections;
            $self->start_timer;
         }
      );
}

sub send_upd_presence {
   my ($self, $contact) = @_;

   if ($self->{xmpp}->get_account ($contact->jid)) {
      return;
   }

   my %status_map = (
      away => 'away',
      dnd  => 'dnd',
      xa   => 'xa',
   );

   my $pres = $self->{xmpp}->get_priority_presence_for_jid ($contact->jid);
   my $sta =
      $pres
         ? $status_map{$pres->show} || 'present'
         : 'n/a';

   if (!$pres && $contact->subscription eq 'remove') {
      $sta = 'dead';
   }

   $self->{server}->change_state (
      $self->to_xmpp_url ($contact->jid),
      $sta,
      ($pres ? ($pres->status || '') : ''),
      $contact->nickname
   );

   if ($sta ne 'dead') {
      my %groups;
      for my $c ($self->{xmpp}->get_contacts_for_jid ($contact->jid)) {
         $groups{$_} = 1 for $c->groups;
      }
      $self->{server}->set_group ($self->to_xmpp_url ($contact->jid), keys %groups);
   }
}

sub send_contact_update {
   my ($self, $id, $cb, $okmsg, %upd) = @_;
   my $errors;
   my @contacts = $self->{xmpp}->get_contacts_for_jid ($self->from_xmpp_url ($id));
   my $set_cnt = scalar @contacts;

   for my $c (@contacts) {
      $c->send_update (
         sub {
            my ($c, $error) = @_;
            if ($error) { $errors .= "Error for $id: " . $error->string }
            $set_cnt--;
            if ($set_cnt <= 0) {
               if ($errors) {
                  $cb->(0, $errors);
               } else {
                  $cb->(1, $okmsg);
               }
            }
         },
         %upd
      );
   }
}

sub set_nick {
   my ($self, $id, $nick, $cb) = @_;
   my $jid = $self->from_xmpp_url ($id);
   if ($self->{xmpp}->get_account ($jid)) {
      $self->{cfg}->{nick} = $nick;
      $self->update_account_presences;
   }
   $self->send_contact_update ($id, $cb, "nickname for $id set to $nick", name => $nick);
}

sub set_group {
   my ($self, $id, $groups, $cb) = @_;
   $self->send_contact_update ($id, $cb, "groups set for $id", groups => $groups);
}

sub set_presence {
   my ($self, $state, $desc) = @_;

   unless (grep { $state eq $_ } qw/dnd away xa chat/) {
      $state = undef;
   }
   # FIXME: 10 is the hardcoded default priority for bummskraut :)
   $self->{xmpp}->set_presence ($state, $desc, 10);
   $self->{last_presence} = [$state, $desc];
   $self->update_account_presences;
}

sub init {
   my ($self) = @_;
   $self->{xmpp}->reg_cb (
      debug_send => sub {
         my ($xmpp, $acc, $data) = @_;
         my $log = _dumpxml ($data);
         $log = "XMPP SEND (".$acc->jid."):\n$log\n";
         $self->{server}->debug_log ($log);
      },
      debug_recv => sub {
         my ($xmpp, $acc, $data) = @_;
         my $log = _dumpxml ($data);
         $log = "XMPP RECV (".$acc->jid."):\n$log\n";
         $self->{server}->debug_log ($log);
      },
      message => sub {
         my ($xmpp, $acc, $msg) = @_;
         my $src = $self->to_xmpp_url ($msg->from);

         my $msgbody = $msg->any_body;

         $self->{server}->distribute_message (
            $src, $src, 'private',
            'normal',
            0, 0, time, $msgbody
         ) if defined $msgbody;
      },
      session_ready => sub {
         my ($xmpp, $acc) = @_;

         $self->{server}->distribute_info (
            undef, xmpp_connect => "Connected to XMPP Account " . $acc->bare_jid
         );

         my ($state, $desc) = @{$self->{last_presence} || ['present', '']};
         $self->{xmpp}->set_presence ($state, $desc, 10);

         $self->update_presences;
      },
      roster_update => sub {
         my ($xmpp, $acc, $roster, $upd) = @_;
         $self->send_upd_presence ($_) for @$upd;
      },
      error => sub {
         my ($xmpp, $acc, $error) = @_;
         $self->{server}->distribute_error (
            undef, xmpp_error => $acc->bare_jid . ": " . $error->string
         );
      },
      presence_update => sub {
         my ($xmpp, $acc, $roster, $contact, $old_pres, $new_pres) = @_;
         $self->send_upd_presence ($contact);
      },
      contact_did_unsubscribe => sub {
         my ($xmpp, $acc, $roster, $contact, $rdoit) = @_;
         $self->{server}->distribute_meta (
            $self->to_xmpp_url ($contact->jid), $self->to_xmpp_url ($contact->jid),
            "unsubscribed"
         );
         1
      },
      contact_subscribed => sub {
         my ($xmpp, $acc, $roster, $contact) = @_;
         $self->{server}->distribute_meta (
            $self->to_xmpp_url ($contact->jid), $self->to_xmpp_url ($contact->jid),
            "contact subscribed to your presence."
         );
      },
      contact_unsubscribed => sub {
         my ($xmpp, $acc, $roster, $contact) = @_;
         $self->{server}->distribute_meta (
            $self->to_xmpp_url ($contact->jid), $self->to_xmpp_url ($contact->jid),
            "contact unsubscribed or declied your subscription request."
         );
      },
      # FIXME: implement checkmechanism for subscription requests!
      contact_request_subscribe => sub {
         my ($xmpp, $acc, $roster, $contact) = @_;
         $contact->send_subscribe;
         1
      },
      disconnect => sub {
         my ($xmpp, $acc, $h, $p, $msg) = @_;
         $self->{server}->distribute_error (
            undef, xmpp_disconnect => $acc->bare_jid . ": disconnected $h:$p: $msg"
         );

         $self->update_presences;
      },
   );

   $self->update_connections;
}

sub cfg_keys {
   my ($self) = @_;

   (
      [jids => "Your JID accounts"],
      [via  => "Connect to a JID via a host:port"],
      [nick => "Your nickname"],
      [main_jid => "Your primary JID"],
      [update_interval => "The interval in seconds to update the connection states and retry connects"],
   )
}

sub get_main_jid_account {
   my ($self, $errmsg) = @_;

   unless ($self->{cfg}->{main_jid}) {
      $self->{server}->distribute_error (
         undef,
         xmpp_no_main_jid => "You didn't configure a main JID yet: $errmsg"
      );
      return;
   }

   my $acc = $self->{xmpp}->get_account ($self->{cfg}->{main_jid});
   unless ($acc) {
      $self->{server}->distribute_error (
         undef,
         xmpp_no_main_jid_connected => "Main JID isn't connected: $errmsg"
      );
      return
   }

   $acc
}

sub subscribe {
   my ($self, $id, $cb) = @_;

   my $acc = $self->get_main_jid_account ("subscribe failed");
   if ($acc) {
      my $cjid = $self->from_xmpp_url ($id);
      my $ros = $acc->connection ()->get_roster ();
      $ros->new_contact ($cjid, undef, undef, sub {
         my $con = $ros->get_contact ($cjid);
         if ($con) {
            $con->send_subscribe;
         } else {
            $cb->(0, "subscribe failed, contact couldn't be created on roster");
         }
      });
      $cb->(1, "subscribe request sent.");
   } else {
      $cb->(0, "subscribe failed, couldn't get main JID account.");
   }
}

sub unsubscribe {
   my ($self, $id, $cb) = @_;

   my $jid = $self->from_xmpp_url ($id);
   for my $ac ($self->{xmpp}->get_connected_accounts) {
      my $ros = $ac->connection ()->get_roster ();
      my $con = $ros->get_contact ($jid);
      next unless $con;
      $con->send_unsubscribe;
      $con->send_unsubscribed;
      $ros->delete_contact ($jid);
   }

   $cb->(1, "unsubscribe sent.");
}

sub subscriptions {
   my ($self) = @_;
   @{$self->{cfg}->{subscriptions}}
}

sub update_account_presences {
   my ($self) = @_;

   for my $ac ($self->{xmpp}->get_accounts) {
      my ($state, $desc) = @{$self->{last_presence} || ['present', '']};
      $state = $state eq '' ? 'present' : $state;

      if (!$ac->is_connected) {
         $state = 'n/a';
      }

      $desc ||= '';

      $self->{server}->change_state (
         $self->to_xmpp_url ($ac->jid), $state, $desc, $self->{cfg}->{nick}
      );
      $self->{server}->set_group ($self->to_xmpp_url ($ac->jid), "*local account*");
   }
}

sub update_presences {
   my ($self) = @_;

   my (@ids) = $self->{server}->get_state_ids;

   for my $id (grep /^xmpp:~/, @ids) {
      my $jid = $self->from_xmpp_url ($id);

      next if $self->{xmpp}->get_account ($jid);

      my (@cons) = $self->{xmpp}->get_contacts_for_jid ($jid);
      if (@cons) {
         $self->send_upd_presence ($cons[0]);
      } else {
         $self->{server}->change_state ($id, 'dead');
      }
   }

   $self->update_account_presences;
}

sub update_connections {
   my ($self) = @_;

   for my $jid (keys %{$self->{cfg}->{jids}}) {
      my @args;
      if (my $hp = $self->{cfg}->{via}->{$jid}) {
         my ($host, $port) = split /:/, $hp, 2;
         push @args, ($host ne '' ? $host : undef);
         push @args, ($port ne '' ? $port : undef);
      }
      $self->{xmpp}->add_account ($jid, $self->{cfg}->{jids}->{$jid}, @args);
   }

   $self->{xmpp}->update_connections
}

sub reload {
   my ($self) = @_;

   $self->{xmpp}->remove_accounts ("reload");

   $self->{reload_timer} =
      AnyEvent->timer (after => 10, cb => sub {
         $self->update_connections;
         delete $self->{reload_timer}
      });
}

sub to_xmpp_url {
   my ($self, $jid) = @_;
   $jid = prep_bare_jid $jid;
   "xmpp:~$jid" # FIXME
}

sub from_xmpp_url {
   my ($self, $id) = @_;
   $id =~ s/^xmpp:~//;
   $id
}

sub cmp_id {
   my ($self, $id1, $id2) = @_;
   lc ($id1) eq lc ($id2) # FIXME
}

# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXX

sub get_src {
   my ($self, $dest) = @_;
   my $acc = $self->get_src_acc ($dest);
   $self->to_xmpp_url ($acc->bare_jid)
}

sub get_src_acc {
   my ($self, $dest) = @_;
   my $jid = $self->from_xmpp_url ($dest);
   $self->{xmpp}->find_account_for_dest_jid ($jid);
}

sub message {
   my ($self, $dest, $scope, $type, $content) = @_;
   my $acc = $self->get_src_acc ($dest);
   my $src = $self->to_xmpp_url ($acc->bare_jid);

   $self->{server}->distribute_message (
      # FIXME: add public for chat rooms once implemented in xmpp
      $src, $dest, 'private',
      $type, 1, 0, my $ts = time, $content
   );

   eval {
      $self->{xmpp}->send_message ($content, $self->from_xmpp_url ($dest), $acc, 'chat');
   };
   if ($@) {
      $self->{server}->distribute_error (undef, xmpp_msg_send => "Couldn't send message to '$dest': $@");
   }
}

sub raw {
   my ($self, $dest, $msg) = @_;
   # XXX: unimplemented!
}

sub meta {
   my ($self, $dest, $msg) = @_;
   #my $acc = $self->get_src_acc ($dest);
   if ($msg =~ /test/) {
      #$self->{xmpp}->remove_account ($self->{xmpp}->get_account ('elmex@amessage.eu'), 'test disconnect');
   }
}

1;
