package Net::Bummskraut::Server;
use strict;
use Net::Bummskraut::Listener;
use Net::Bummskraut::Util;
use Digest::HMAC_SHA1 qw/hmac_sha1_hex/;

our @ISA = qw/Net::Bummskraut::Listener/;

=pod

Following methods need to be implemented by a scheme:

=over 4

=item subscribe ($dest, $cb)

$cb->($subscribed, $message);

=item unsubscribe ($dest, $cb)

$cb->($unsubscribed, $message);

=item subscriptions ()

=item set_group ($id, $groups, $cb)

=item set_nick ($id, $nick, $cb)

=item set_presence ($state, $desc)

=item message ($dest, $scope, $type, $content)

=item get_src ($dest)

=item raw ($dest, $msg)

=item meta ($dest, $msg)

=item cfg_get ($key) => $value

=item cfg_keys () => @keys => $keys[0] = [$key, $desc]

=item cfg_set ($key, $value, $listval) => $need_reload_bool

=item reload

=item refresh

=back

=cut

sub new {
   my $this = shift;
   my $class = ref($this) || $this;
   my $self = {@_};
   bless $self, $class;
   $self->start_listener;
   $self->reg_cb (connect => \&handle_connect);
   $self->set_cfg_path ($self->{cfg_path});

   $self->start_log_flusher;

   return $self;
}

sub start_log_flusher {
   my ($self) = @_;

   $self->{_lflusher} =
      AnyEvent->timer (after => 20, cb => sub {
         $self->flush_logs;
         $self->start_log_flusher;
      });
}

sub flush_logs {
   my ($self) = @_;
   for (values %{$self->{logfiles}}) {
      $_->flush;
   }
}

sub debug_log {
   my ($self, $data) = @_;
   my @lines = split /\n/, $data;
   $self->log_line (debug_log => time, $_) for @lines;
}

sub log_line {
   my ($self, $id, $time, $data) = @_;

   $id =~ s/\//#/g;

   my $logfh = $self->{logfiles}->{$id};

   unless ($logfh) {
      my $logdir = $ENV{HOME} . "/.bummskraut_server";
      mkdir $logdir unless -e $logdir;
      unless (open $logfh, ">>", "$logdir/$id") {
         $self->distribute_error (undef, log_error => time, "Couldn't open logfile: '$logdir/$id': $!");
         return;
      }
      $self->{logfiles}->{$id} = $logfh;
   }

   my $ts = POSIX::strftime ("%F %T %z", localtime ($time || time));
   $logfh->printf ("%s | %s\n", $ts, $data);
}

sub add_last_log {
   my ($self, $id, $log) = @_;
   push @{$self->{llog}->{$id}}, $log;
}

sub get_last_log {
   my ($self, $id) = @_;
   @{$self->{llog}->{$id} || []}
}

=head2 set_cfg_path ($path)

Sets the config path and loads the config.

=cut

sub set_cfg_path {
   my ($self, $path) = @_;
   $self->{cfg} = load_cfg ($self->{cfg_path} = $path);

   unless (exists $self->{cfg}->{key}) {
      $self->{cfg}->{key} =
         join '', map { unpack ("h*", $self->read_random) } 1..5;

      unless (-e $ENV{HOME}."/.bummskraut_key") {
         open my $bk, ">", $ENV{HOME}."/.bummskraut_key"
            or return;
         binmode $bk, ":UTF-8";
         $bk->print ($self->{cfg}->{key});
      }

      $self->commit_cfg;
   }
}

=head2 print_auth_key

Prints the authentication key to STDOUT.

=cut

sub print_auth_key {
   my ($self) = @_;
   print "Authentication key: $self->{cfg}->{key}\n";
}

=head2 get_my_cfg ($schema)

=cut

sub get_my_cfg {
   my ($self, $key) = @_;
   $self->{cfg}->{$key} = {} unless exists $self->{cfg}->{$key};
   $self->{cfg}->{$key}
}

=head2 commit_cfg

=cut

sub commit_cfg {
   my ($self) = @_;
   save_cfg ($self->{cfg_path}, $self->{cfg});
}

=head2 reg_scheme ($scheme, $handle)

Sets a scheme handler.

=cut

sub reg_scheme {
   my ($self, $scheme, $handle) = @_;
   $self->{schemes}->{$scheme} = $handle;
}

=head2 get_schemes ()

Returns all registered schemes.

=cut

sub get_schemes {
   my ($self) = @_;
   return %{$self->{schemes}}
}

=head2 get_scheme ($name)

Returns the scheme of the name C<$name>

=cut

sub get_scheme {
   my ($self, $s) = @_;
   $self->{schemes}->{$s}
}

sub get_scheme_handler {
   my ($self, $dest) = @_;
   my ($sc, $opq) = split /:/, $dest, 2;
   $self->{schemes}->{$sc};
}

sub for_all_schemes {
   my ($self, $cb) = @_;
   for (values %{$self->{schemes}}) {
      $cb->($self, $_);
   }
}

sub get_handler {
   my ($self, $cl, $packet, $dest) = @_;
   my $schdl = $self->get_scheme_handler ($dest);
   unless ($schdl) {
      $cl->srv_error ($packet, $dest, bad_scheme => time, "Couldn't handle scheme for '$dest'");
      return
   }
   $schdl
}

sub read_random {
   open my $rnd, "/dev/urandom" 
      or do { warn "Couldn't open /dev/urandom: $!"; return "" . time };
   sysread $rnd, my $rnd, 20;
   $rnd
}

sub handle_connect {
   my ($self, $cl) = @_;

   $cl->reg_cb (
      hello   => sub {
         my ($cl, $packet) = @_;
         my $rnd = unpack ("h*", $self->read_random);
         $cl->{_auth_digest} = hmac_sha1_hex ($rnd, $self->{cfg}->{key});
         $cl->srv_auth_challenge ($packet, $rnd);

         0
      },
      auth_response => sub {
         my ($cl, $packet, $response) = @_;
         if ($cl->{_auth_digest} eq $response) {
            $cl->srv_hello ($packet);
            $self->cl_send_states ($cl);
            $self->cl_send_groups ($cl);
            $self->cl_send_subids ($cl);
            $self->cl_send_last_logs ($cl);

            $cl->reg_cb (
               subscribe   => sub { $self->handle_subscribe (@_); 1 },
               unsubscribe => sub { $self->handle_unsubscribe (@_); 1 },
               message     => sub { $self->handle_cl_message (@_); 1 },
               raw_command => sub { $self->get_scheme_handler ($_[2])->raw ($_[2], $_[3]); 1 }, # FIXME: raw reply message make do!
               meta_command=> sub { $self->get_scheme_handler ($_[2])->meta ($_[2], $_[3]); 1 }, # FIXME: raw reply message make do!
               config_get  => sub { $self->handle_cfg_get (@_); 1 },
               config_set  => sub { $self->handle_cfg_set (@_); 1 },
               broadcast_relay => sub { $self->handle_broadcast_relay (@_); 1 },
               command_history => sub { $self->handle_command_history (@_); 1 },
               set_group   => sub { $self->handle_set_group (@_); 1 },
               set_nick    => sub { $self->handle_set_nick (@_); 1 },
               set_presence => sub { $self->handle_set_presence (@_); 1 },
            );

         } else {
            $cl->srv_error ($packet, undef, authentication_failure => time,
               "You are not authorized to use this service, please check whether the key in ~/.bummskraut_key matches the one of the bummskraut_server instance you want to use."
            );
         }

         0
      }
   );
}

sub handle_set_presence {
   my ($self, $cl, $packet, $state, $desc) = @_;

   $self->for_all_schemes (sub {
      my ($self, $scheme) = @_;
      $scheme->set_presence ($state, $desc);
   });
}

sub handle_set_nick {
   my ($self, $cl, $packet, $id, $nick) = @_;
   my $schdl = $self->get_handler ($cl, $packet, $id)
                  or return;

   $schdl->set_nick ($id, $nick, sub {
      my ($succ, $message) = @_;
      if ($succ) {
         $cl->srv_info ($packet, $id, set_nick => time, $message);
      } else {
         $cl->srv_error ($packet, $id, set_nick => time, $message);
      }
   });
}

sub handle_set_group {
   my ($self, $cl, $packet, $id, $groups) = @_;
   my $schdl = $self->get_handler ($cl, $packet, $id)
                  or return;

   $schdl->set_group ($id, $groups, sub {
      my ($succ, $message) = @_;
      if ($succ) {
         $cl->srv_info ($packet, $id, set_groups => time, $message);
      } else {
         $cl->srv_error ($packet, $id, set_groups => time, $message);
      }
   });
}

sub handle_cfg_get {
   my ($self, $cl, $packet, $cfgname) = @_;
   $self->{cfg}->{_clcfg}->{$cfgname} ||= {};
   $cl->srv_config ($packet, $cfgname, $self->{cfg}->{_clcfg}->{$cfgname});
}

sub handle_cfg_set {
   my ($self, $cl, $packet, $cfgname, $cfg) = @_;
   $self->{cfg}->{_clcfg}->{$cfgname} = $cfg;
   $self->commit_cfg;

   $cl->srv_info (
      $packet, undef, config_set => time, "set configuration for client '$cfgname'."
   );

   $self->foreach_client_except ($cl, sub {
      my ($self, $cl) = @_;
      $cl->srv_config (undef, $cfgname, $cfg);
   });
}

sub handle_subscribe {
   my ($self, $cl, $packet, $dest) = @_;
   my $schdl = $self->get_handler ($cl, $packet, $dest)
                  or return;

   $schdl->subscribe ($dest, sub {
      my ($subscribed, $message) = @_;
      if ($subscribed) {
         $cl->srv_info ($packet, $dest, subscribed => time, $message);
      } else {
         $cl->srv_error ($packet, $dest, subscribe => time, $message);
      }
   });
}

sub handle_unsubscribe {
   my ($self, $cl, $packet, $dest) = @_;
   my $schdl = $self->get_handler ($cl, $packet, $dest)
                  or return;

   $schdl->unsubscribe ($dest, sub {
      my ($unsubscribed, $message) = @_;
      if ($unsubscribed) {
         $cl->srv_info ($packet, $dest, unsubscribed => time, $message);
      } else {
         $cl->srv_error ($packet, $dest, unsubscribe => time, $message);
      }
   });
}

sub handle_cl_message {
   my ($self, $cl, $packet, $src, $dest, $scope, $type,
       $is_echo, $do_highlight, $ts, $content)   = @_;

   my $schdl = $self->get_handler ($cl, $packet, $dest)
                  or return;

   $schdl->message ($dest, $scope, $type, $content);
   my $src = $schdl->get_src ($dest)
      or return;

   $cl->srv_message_sent ($packet, $src, $dest, time);
}

sub handle_broadcast_relay {
   my ($self, $cl, $packet, $command, $tag, $data, $ts) = @_;

   $cl->srv_broadcast_relay ($packet, $command, $tag, $data, $ts);

   $self->foreach_client_except ($cl, sub {
      my ($self, $cl) = @_;
      $cl->srv_broadcast_relay (undef, $command, $tag, $data, $ts);
   });
}

sub cl_send_states {
   my ($self, $cl) = @_;

   for (keys %{$self->{states}}) {
      my $s = $self->{states}->{$_};
      $cl->srv_state_change ($_, time, $s->[0], $s->[1], $s->[2]);
   }
}

sub get_src {
   my ($self, $id) = @_;
   my $schdl = $self->get_scheme_handler ($id);
   unless ($schdl) {
      warn "fatal: no scheme handler for: $id, ";
      return;
   }
   $schdl->get_src ($id);
}

sub send_last_log {
   my ($self, $cl, $type, @args) = @_;

   if ($type eq 'meta') {
      $cl->srv_meta_message (@args)

   } elsif ($type eq 'msg') {
      $cl->srv_message (undef, @args)
   }
}

sub handle_command_history {
   my ($self, $cl, $packet, $id, $lines) = @_;
   $lines ||= 20;
   my (@log) = $self->get_last_log ($id);

   my (@lines) = (@log > $lines ? splice @log, -$lines : @log);
   $cl->srv_info ($packet, $id, history_sent => time, "Sending ".scalar (@lines)." lines of history for $id...");
   for (@lines) {
      $self->send_last_log ($cl, @$_);
   }
}

sub cl_send_last_logs {
   my ($self, $cl) = @_;

   for (keys %{$self->{llog}}) {
      my (@log) = $self->get_last_log ($_);
      for (@log > 20 ? splice @log, -20 : @log) {
         $self->send_last_log ($cl, @$_);
      }
   }
}

=head2 change_state ($id, $state, $desc, $alias)

Sets the C<$id>s state.

=cut

sub change_state {
   my ($self, $id, $state, $desc, $alias) = @_;

   my $update = 0;

   if ($state eq 'dead') {
      $update = delete $self->{states}->{$id};
      delete $self->{subids}->{$id};
      $desc ||= '';
      $alias ||= '';
   } else {
      my $prev = $self->{states}->{$id};
      if ($prev->[0] ne $state
          || $prev->[1] ne $desc
          || ((defined $alias) && $prev->[2] ne $alias))
      {
         $update = 1;
      }
      $self->{states}->{$id} = [$state, $desc, $alias];
   }

   if ($update) {
      my $ts = time;
      $self->foreach_client (sub {
         my ($self, $cl) = @_;
         $cl->srv_state_change ($id, $ts, $state, $desc, $alias);
      });
      $self->log_line ($id, $ts, "$id (aka $alias) changed state to $state ($desc)");
   }
}

sub get_state_ids {
   my ($self) = @_;
   return keys %{$self->{states}};
}

sub cl_send_subids {
   my ($self, $cl) = @_;

   for my $id (keys %{$self->{subids}}) {
      $cl->srv_add_subid ($id, time, values %{$self->{subids}->{$id}});
   }
}

=head2 remove_subids ($id, @subids)

C<@subids> is a list of tuples: (id, nick)

=cut

sub remove_subids {
   my ($self, $id, @subids) = @_;

   for (@subids) {
      my $sub = delete $self->{subids}->{$id}->{$_->[0]};
      $self->log_line ($id, time, "- $sub->[0] (aka $sub->[1])");
   }

   $self->foreach_client (sub {
      my ($self, $cl) = @_;
      $cl->srv_remove_subid ($id, time, @subids);
   });
}

=head2 add_subids ($id, @subids)

C<@subids> is a list of tuples: (id, nick)

=cut

sub add_subids {
   my ($self, $id, @subids) = @_;

   for (@subids) {
      $self->{subids}->{$id}->{$_->[0]} = $_;
      $self->log_line ($id, time, "+ $_->[0] (aka $_->[1])");
   }

   $self->foreach_client (sub {
      my ($self, $cl) = @_;
      $cl->srv_add_subid ($id, time, @subids);
   });
}

=head2 change_subid ($id, $subid, $newsubid, $oldnick, $newnick)

A subid changes, either it's source and/or it's nickname.

=cut

sub change_subid {
   my ($self, $id, $subid, $newsubid, $oldnick, $newnick) = @_;

   delete $self->{subids}->{$id}->{$subid};
   $self->{subids}->{$id}->{$newsubid} = [$newsubid, $newnick];

   $self->log_line ($id, time, "% $subid (aka $oldnick) => $newsubid (aka $newnick)");

   $self->foreach_client (sub {
      my ($self, $cl) = @_;
      $cl->srv_change_subid ($id, time, $subid, $newsubid, $oldnick, $newnick);
   });
}

sub get_id_alias {
   my ($self, $id, $dest) = @_;
   my $nick;

   if (defined $dest) {
      if (exists $self->{subids}->{$dest}
          && exists $self->{subids}->{$dest}->{$id})
      {
         $nick = $self->{subids}->{$dest}->{$id}->[1];
      }
   }

   unless (defined $nick) {
      $nick = $self->{states}->{$id}->[2]
         if exists $self->{states}->{$id};
   }
   $nick || $id
}

=head2 distribute_error ($src, $type, $content)

=cut

sub distribute_error {
   my ($self, $src, $type, $content) = @_;

   my $ts = time;
   $self->log_line ($src || 'status', $ts, "ERROR($type): $content");

   $self->foreach_client (sub {
      my ($self, $cl) = @_;
      $cl->srv_error (
         undef, $src, $type, $ts, $content
      );
   });
}

=head2 distribute_info ($src, $type, $content)

=cut

sub distribute_info {
   my ($self, $src, $type, $content) = @_;

   my $ts = time;
   $self->log_line ($src || 'status', $ts, "info($type): $content");

   $self->foreach_client (sub {
      my ($self, $cl) = @_;
      $cl->srv_info (
         undef, $src, $type, $ts, $content
      );
   });
}

=head2 distribute_meta ($src, $dest, $content)

=cut

sub distribute_meta {
   my ($self, $src, $dest, $content) = @_;

   my $ts = time;

   if ($src) {
      my $srcn = $self->get_id_alias ($src, $dest);
      $self->log_line ($dest, $ts, "meta [from $src (aka $srcn)]: $content");
   }

   $self->add_last_log ($dest, [meta => $src, $dest, $ts, $content]);

   $self->foreach_client (sub {
      my ($self, $cl) = @_;
      $cl->srv_meta_message (
         $src, $dest, $ts, $content
      );
   });
}

=head2 distribute_message ($src, $dest, $scope, $type, $is_echo, $do_highlight, $ts, $content)

=cut

sub distribute_message {
   my ($self, $src, $dest, $scope, $type,
       $is_echo, $do_highlight, $ts, $content) = @_;

   { # logging
      my $line = sprintf (
         "[%-7s/%-7s] %1s <%s> %s",
         $scope, $type, ($is_echo ? 'E' : ''),
         $self->get_id_alias ($src, $dest), $content
      );
      $self->log_line ($dest, $ts, $line);
   }

   $self->add_last_log ($dest, [msg => $src, $dest, $scope, $type, $is_echo, $do_highlight, $ts, $content]);

   $self->foreach_client (sub {
      my ($self, $cl) = @_;
      $cl->srv_message (
         undef,
         $src, $dest, $scope, $type,
         $is_echo, $do_highlight, $ts, $content
      );
   });
}

=head2 set_group ($id, @groups)

=cut

sub set_group {
   my ($self, $id, @groups) = @_;
   $self->{groups}->{$id} = [ @groups ];

   $self->foreach_client (sub {
      my ($self, $cl) = @_;
      $cl->srv_set_group ($id, $self->{groups}->{$id});
   });
}

sub cl_send_groups {
   my ($self, $cl) = @_;

   for (keys %{$self->{groups}}) {
      my $s = $self->{groups}->{$_};
      $cl->srv_set_group ($_, $s);
   }
}
